--[[
------------------------------------------------------------------------
	ListBox:DMAdd(entry [,state [,indent [,icon] ] ] )
		Add entry to the end of the list
	ListBox:DMSet(entry [,state [,indent [,icon [,position] ] ] ] )
		Set/change an existing entry

	entry   : Visible text
	state   : DM_STATE_LIST|DM_STATE_INACTIVE|DM_STATE_CHECKED|DM_STATE_UNCHECKED
	indent  : 0|1|2...
	icon    : file-name
	position: Index in list to change

------------------------------------------------------------------------
	ListBox:DMClear()

	Clear the list.

------------------------------------------------------------------------
	Events

	The events are sent to the ListBox parent with these parameters,
	excluding the hidden "self":
		1: Handle of the list-item widget
		2: Event (as listed below)
		3: Handle of the ListBox frame
		4: The index-number the clicked item represents in the list


	DMEVENT_LISTBOX_ITEM_CLICKED
		Sent when a list-item has been clicked
	DMEVENT_LISTBOX_ITEM_ENTER/LEAVE
		Sent when the mouse enters/leaves a list-item

------------------------------------------------------------------------
	XML usage

	<Frame name="$parent_ListBox" inherits="DuckMod_ListBox_01">
		<Size><AbsDimension x="245" y="109"/></Size>
		<Anchors>
			<Anchor point="TOPLEFT"><Offset><AbsDimension x="11" y="-125" /></Offset></Anchor>
		</Anchors>
	</Frame>

------------------------------------------------------------------------
	Lua usage

	The ListBox's parent needs an OnEvent handler, which will be called
	with the above mentioned events.


========================================================================
========================================================================
IMPORTANT
========================================================================
The following scripts are reserved for the ListBox widget. If you
desperately need some of these, hook them in Lua in stead of defining
them in XML.
	OnLoad
	OnMouseWheel
	OnScrollRangeChanged
========================================================================
]]


DuckMod_Widgets_Version=DuckMod_Widgets_Version or 0;


if (DuckMod_Widgets_Version<2.03) then			-- This version
	DuckMod_Widgets_Version=2.03;

	        DM_STATE_LIST=0x0000;	-- Just dumb text
	    DM_STATE_INACTIVE=0x0001;	-- Unclickable button
	     DM_STATE_CHECKED=0x0002;	-- Checked button v1 (exclusive)
	  DM_STATE_CHECKEDexc=0x0002;	-- Checked button v2 exclusive
	   DM_STATE_UNCHECKED=0x0003;	-- Unchecked button v1 (exclusive)
	DM_STATE_UNCHECKEDexc=0x0003;	-- Unchecked button v2 exclusive
	      DM_STATE_MASKv1=0x0FFF;	-- Mask for testing v1 compatibility
	      DM_STATE_MASKv2=0xF000;	-- Mask for testing v2 settings
	  DM_STATE_CHECKEDinc=0x8002;	-- Checked button v2 inclusive
	DM_STATE_UNCHECKEDinc=0x8003;	-- Unchecked button v2 inclusive

	if (not DuckWidget) then DuckWidget={}; end	-- If it exists, just overlay
	if (not DuckWidget.ListBox) then DuckWidget.ListBox={}; end	-- If it exists, just overlay
	if (not DuckWidget.CheckBox) then DuckWidget.CheckBox={}; end	-- If it exists, just overlay


function DuckWidget:CopyTable(t)
	if (not t) then return nil; end
	local new={};
	local i,v;
	for i,v in pairs(t) do
		if (type(v)=="table") then new[i]=self:CopyTable(v); else new[i]=v; end
	end
	return new;
end


-- Generic mouse-wheel handler for scrolling-panes.
-- o The main use for this is when a pane has other items in it that
--   may receive it in stead. They may then call this function with
--   their parent as an argument.
function DuckWidget.MouseWheel(slider,delta)
	if (not slider) then return; end
	local min,max=slider:GetMinMaxValues(); if (not min or not max) then return; end
	local step=slider:GetValueStep(); if (not step) then return; end
	local value=slider:GetValue(); if (not value) then return; end
	if (delta==1 and (value-step)>=min) then slider:SetValue(value-step);
	elseif (delta==-1 and (value+step)<=max) then slider:SetValue(value+step); end
end

-- Some helpers to get past later updates to the WoW code
function DuckWidget.CheckBox.SetButtonText(button,text,subtext) if (text) then _G[button:GetName().."Text"]:SetText(text); end if (subtext) then _G[button:GetName().."SubText"]:SetText(subtext); end end
function DuckWidget.CheckBox.GetButtonText(button) return _G[button:GetName().."Text"]:GetText(); end
function DuckWidget.CheckBox.SetButtonTextColor(button,r,g,b,a) _G[button:GetName().."Text"]:SetTextColor(r,g,b,a); end
-- CheckBox
function DuckWidget.CheckBox.SetIcon(button,image) _G[button:GetName().."IconTexture"]:SetTexture(image); end
function DuckWidget.CheckBox.GetSize(button) return button:GetHeight(); end
function DuckWidget.CheckBox.SetSize(button,size)
	button.defaultIndent=size+ceil(size*0.1);	-- Add about 10% spacing
	button:SetHeight(size);
	_G[button:GetName().."IconTexture"]:SetWidth(size);
	_G[button:GetName().."IconTexture"]:SetHeight(size);
end

-- ListBox
function DuckWidget.ListBox.VisibleEntries(frame)
	local fHeight=frame:GetHeight();
	if (not _G[frame:GetName().."_Entry1"]) then return 1; end
	local eHeight=_G[frame:GetName().."_Entry1"]:GetHeight();
	return floor((fHeight+frame.DMSpacing)/(eHeight+frame.DMSpacing));
end

function DuckWidget.ListBox.Clear(frame)
	frame.DMEntries=0;
	_G[frame:GetName().."_Scroll"]:SetMinMaxValues(1,1);		-- Set span
	_G[frame:GetName().."_Scroll"]:SetValue(1);
	DuckWidget.ListBox.Redraw(frame);
end

-- Add an entry to the list
-- self    : The frame
-- entry   : Text of the entry to add
-- state   : LIST, checked, unchecked, inactive
-- position: END, index
function DuckWidget.ListBox.AddDouble(frame,entry,subtext,state,indent,icon) return DuckWidget.ListBox.Add(frame,entry.."\n"..subtext,state,indent,icon) end
function DuckWidget.ListBox.Add(frame,entry,state,indent,icon)
	if (not entry) then return; end
	local position,subtext=nil,nil;
	if (entry:find("§",1,true)) then
		local buf={strsplit("§",entry)};
		for _,entry in ipairs(buf) do if (strtrim(entry)~="") then DuckWidget.ListBox.Add(frame,entry,state,indent,icon); end end
		return;
	end
	frame.DMEntries=frame.DMEntries+1;									-- One more in list
	local length=(frame.DMEntries-DuckWidget.ListBox.VisibleEntries(frame))+1;
	if (length<1) then length=1; end
	local _,cur=_G[frame:GetName().."_Scroll"]:GetMinMaxValues();	-- Get span
	if (cur~=length) then _G[frame:GetName().."_Scroll"]:SetMinMaxValues(1,length); end		-- Set span
	if (not position) then position=frame.DMEntries; end	-- Position not provided, so add at the end
	local i=frame.DMEntries;								-- Make room for it if it's not appended
	while(i>position) do frame.DMTheList[i]=frame.DMTheList[i-1]; i=i-1; end
	return DuckWidget.ListBox.Set(frame,entry,state,indent,icon,position);	-- Change it to the supplied info
end

-- Change an entry in the list
-- self    : The frame
-- entry   : Text of the entry to set
-- state   : LIST, checked, unchecked, inactive
-- position: END, index
function DuckWidget.ListBox.Set(frame,entry,state,indent,icon,position)
	if (not entry) then entry=""; end
	if (not position) then return DuckWidget.ListBox.Add(frame,entry,state,indent,icon); end
	if (not frame.DMTheList) then frame.DMTheList={}; end
	if (not state) then state=DM_STATE_LIST; end
	if (not indent) then indent=0; end
	local entry={strsplit("\n",entry)};
	frame.DMTheList[position]={ Entry=entry[1], SubText=entry[2], State=state, Tooltip=nil, Indent=indent, Icon=icon, DB={}, };
	if (frame.defaultSort) then position=DuckWidget.ListBox.Sort(frame.DMTheList,frame.DMEntries,position); end
	DuckWidget.ListBox.Redraw(frame);
	return frame.DMTheList[position];
end

function DuckWidget.ListBox.Sort(list,length,index)
	indent=list[index].Indent;		-- This level
	-- Up
	while(index>1 and list[index-1].Indent==indent and tostring(list[index].Entry)<tostring(list[index-1].Entry)) do
		list[index-1],list[index]=list[index],list[index-1];
		index=index-1;
	end
	-- Down
	while(index<length and list[index+1].Indent==indent and tostring(list[index].Entry)>tostring(list[index+1].Entry)) do
		list[index+1],list[index]=list[index],list[index+1];
		index=index+1;
	end

	return index;
end

function DuckWidget.ListBox.Redraw(frame,remake)
	if (frame.DMEntries==nil) then return; end
	if (not frame.DMTheList) then frame.DMTheList={}; end			-- No list
	local fName=frame:GetName();
	local bName=fName.."_Entry";
	local bSize=frame.DMItemHeight or 17;
	if (_G[bName.."1"]:GetSize()~=bSize) then remake=true; end	-- force remake
	_G[bName.."1"]:SetSize(bSize);								-- set first size for visible calculation
	local i;
	local visible=DuckWidget.ListBox.VisibleEntries(frame);

	-- Visible area not populated by buttons, so create all buttons
	if (not _G[fName.."_Entry"..visible] or remake) then
		-- Create and place all buttons
		i=1;
		while (i<=visible) do
			local button=_G[bName..i];
			if (not button) then button=CreateFrame("CheckButton",bName..i,frame,"DuckMod_CheckBoxA_01"); end
			if (i>1) then					-- never reposition topmost button
				button:ClearAllPoints();
				button:SetSize(bSize);
				button:SetPoint("TOPLEFT",_G[bName..(i-1)],"BOTTOMLEFT",0,-frame.DMSpacing);
				button:SetPoint("RIGHT",_G[bName..(i-1)],"RIGHT");
			end
			button:SetCheckedTexture("Interface\\Buttons\\UI-Listbox-Highlight2","ADD");
			button:GetCheckedTexture():SetVertexColor(0,0,1,nil);	-- make blue checked textures
			button:Hide();
			i=i+1;
		end
		while(_G[bName..i]) do _G[bName..i]:Hide(); i=i+1; end	-- Hide unused buttons due to size-changes
	end

	-- Update from first visible entry
	local offset=_G[fName.."_Scroll"]:GetValue()-1;
	if (offset<0) then offset=0; end
	i=1;
	while(i<=visible and i<=frame.DMEntries-offset) do
		local button=_G[bName..i];
		local entry=frame.DMTheList[offset+i];
		if (entry) then
			button.DMHaveSubText=(bSize>=32);
			local indent=((entry.Indent or 0)+1)*button.defaultIndent;
			local mtxt=_G[bName..i.."Text"];
			local stxt=_G[bName..i.."SubText"];
			mtxt:ClearAllPoints();
			mtxt:SetPoint("LEFT",button,"LEFT",indent,0);
			mtxt:SetPoint("RIGHT",button,"RIGHT",-5,0);
			if (not button.DMHaveSubText) then
				local fi,si,fl=mtxt:GetFont();
--				if (si~=button.defaultTextHeight) then mtxt:SetFont(fi,button.defaultTextHeight,fl); end
				if (si~=bSize*0.95) then mtxt:SetFont(fi,bSize*0.95,fl); end
				stxt:Hide();
			else
				local ms=floor(((bSize/3)*2)-(bSize/33));	-- 2/3-3%
				local ss=floor((bSize/3)-(bSize/33));		-- 1/3-3%
				mtxt:SetPoint("TOP",button,"TOP",0,0);
				local fi,si,fl=mtxt:GetFont();
				if (si~=ms) then mtxt:SetFont(fi,ms,fl); end
				stxt:ClearAllPoints();
				stxt:SetPoint("LEFT",button,"LEFT",indent,0);
				stxt:SetPoint("RIGHT",button,"RIGHT",-5,0);
				stxt:SetPoint("BOTTOM",button,"BOTTOM",0,0);
				fi,si,fl=stxt:GetFont();
				if (si~=ss) then stxt:SetFont(fi,ss,fl); end
				stxt:Show();
			end
			button:SetText(entry.Entry or "",entry.SubText or "");
			button:SetIcon(entry.Icon);
			button.ListIndex=offset+i;
			if (bit.band(entry.State,DM_STATE_MASKv1)==DM_STATE_CHECKEDexc) then button:SetChecked(true); else button:SetChecked(nil); end
			button:Show();
		else button:Hide(); end
		i=i+1;
	end
	while(i<=visible) do _G[bName..i]:Hide(); i=i+1; end	-- Hide unused buttons
end

function DuckWidget.ListBox.SliderChanged(frame) DuckWidget.ListBox.Redraw(frame); end

function DuckWidget.ListBox.SendEvent(button,event,...)
	local parent=button:GetParent():GetParent();
	local handler=parent:GetScript("OnEvent");
	if (not handler) then
		parent=button:GetParent();
		handler=parent:GetScript("OnEvent");
		if (not handler) then return; end
	end
	handler(parent,event,button:GetParent(),button.ListIndex,button:GetChecked(),...);
end

function DuckWidget.ListBox.ListButtonClicked(button,key)
	local i=button:GetName():match("_Entry(%d+)"); if (not i) then return; end	-- unknown button
	local parent=button:GetParent();
	local index=(_G[parent:GetName().."_Scroll"]:GetValue()-1)+i;
	local state=parent.DMTheList[index].State;

	if (state==DM_STATE_CHECKEDinc) then state=DM_STATE_UNCHECKEDinc;
	elseif (state==DM_STATE_UNCHECKEDinc) then state=DM_STATE_CHECKEDinc;
	elseif (state==DM_STATE_CHECKEDexc) then state=DM_STATE_UNCHECKEDexc;		-- v1 compatible
	elseif (state==DM_STATE_UNCHECKEDexc) then state=DM_STATE_CHECKEDexc;	-- v1 compatible
		-- Set all exclusive buttons (radio) unchecked
		local tmp=1;
		while (tmp<=parent.DMEntries) do
			if (parent.DMTheList[tmp].State==DM_STATE_CHECKEDexc) then
				parent.DMTheList[tmp].State=DM_STATE_UNCHECKEDexc;
			end
			tmp=tmp+1;
		end
	end

	-- Set correct for this button
	parent.DMTheList[index].State=state;

	DuckWidget.ListBox.Redraw(parent);
	DuckWidget.ListBox.SendEvent(button,"DMEVENT_LISTBOX_ITEM_CLICKED",key);
end

function DuckWidget.ListBox.ListButtonHover(button,enter)
	local event="DMEVENT_LISTBOX_ITEM_";
	if (enter) then event=event.."ENTER"; else event=event.."LEAVE"; end
	DuckWidget.ListBox.SendEvent(button,event);
end

function DuckWidget.ListBox.SetFuncs(frame)
	frame.DMAdd=DuckWidget.ListBox.Add;
	frame.DMAddDouble=DuckWidget.ListBox.AddDouble;
	frame.DMSet=DuckWidget.ListBox.Set;
	frame.DMClear=DuckWidget.ListBox.Clear;
	frame.DMUpdate=DuckWidget.ListBox.Redraw;
end


end		-- DuckMod_Widgets_Version < this version
